:- module(mini_nat, [{}/1]).

% {Korlat}: Korlat fennll.
{Korlat} :-
        korlat_cel(Korlat, Cel), call(Cel).

		     
% plusz(X, Y, Z): X+Y=Z (Peano szmokkal).
:- block plusz(-, ?, -).
plusz(0, Y, Y).
plusz(s(X), Y, s(Z)) :-
        plusz(X, Y, Z).

% +(X, Y, Z): X+Y=Z (Peano szmokkal). Hatkonyabb, mert
% tovbblp, ha brmelyik argumentum behelyettestett.
:- block +(-, -, -).
+(X, Y, Z) :- 
        var(X), !, plusz(Y, X, Z).  % \+((var(Y),var(Z)))
+(X, Y, Z) :-
        /* nonvar(X), */ plusz(X, Y, Z).

% X-Y=Z, ahol X, Y s Z Peano szmok.
-(X, Y, Z) :-
        +(Y, Z, X).

% X*Y=Z, ahol X, Y s Z Peano szmok. Blokkol, ha nincs tmr argumentuma.
*(X, Y, Z) :-
        when( (ground(X);ground(Y);ground(Z)),
              szorzat(X, Y, Z)
            ).

% X*Y=Z, ahol X, Y s Z Peano szmok. Legalbb egyik argumentuma tmr.
szorzat(X, Y, Z) :-
        (   ground(X) -> szor(X, Y, Z)
        ;   ground(Y) -> szor(Y, X, Z)
        ;   /* Z tmr! */
            Z == 0 -> szorzatuk_nulla(X, Y)
        ;   +(X, _, Z), % X =< Z
            szor(X, Y, Z)
        ).

% X*Y=0, ahol X s Y Peano szmok.
szorzatuk_nulla(X, Y) :-
        (   X = 0 
        ;   dif(X, 0), Y = 0
        ).

% A fenti eljrs korbbi, nem tkletes vltozatai:
% szorzatuk_nulla(X, Y) :-
%        ( X = 0 ; Y = 0 ).
% szorzatuk_nulla(X, Y) :-
%        ( X = 0 ; X \== Y, Y = 0 ).



% szor(X, Y, Z): X*Y=Z, ahol X, Y s Z peano szmok. X tmr.
szor(0, _X, 0).
szor(s(X), Y, Z) :-
        +(Z1, Y, Z),
        szor(X, Y, Z1).

% {Korlat}: Korlat fennll.
{Korlat} :-
        korlat_cel(Korlat, Cel), call(Cel).

% korlat_cel(Korlat, Cel): Korlat lefordtott alakja Cel.
korlat_cel((K1,K2), C12) :-
        korlat_cel(K1, C1), korlat_cel(K2, C2),
	osszetett_cel(C1, C2, C12).
korlat_cel(Kif1=Kif2, C12) :-
        kiertekel(Kif1, E, C1), kiertekel(Kif2, E, C2),
	osszetett_cel(C1, C2, C12).
korlat_cel(Kif1 =< Kif2, Cel) :-
        korlat_cel(Kif1+_ = Kif2, Cel).
korlat_cel(Kif1 < Kif2, Cel) :-
        korlat_cel(s(Kif1) =< Kif2, Cel).
korlat_cel(Kif1 >= Kif2, Cel) :-
        korlat_cel(Kif2 =< Kif1, Cel).
korlat_cel(Kif1 > Kif2, Cel) :-
        korlat_cel(Kif2 < Kif1, Cel).

% kiertekel(Kif, E, Cel): A Kif aritmetikai kifejezs rtkt E-ben
% elllt cl Cel. Kif egszekbl a +, -, s * opertorokkal pl fel.
kiertekel(Kif, E, C123) :-
        compound(Kif), Kif =.. [Op,Kif1,Kif2], !,
        kiertekel(Kif1, E1, C1), kiertekel(Kif2, E2, C2),
        Rel =.. [Op,E1,E2,E],
	osszetett_cel(C1, C2, C12),
	osszetett_cel(C12, Rel, C123).
kiertekel(N, Kif, true) :-
        number(N), !, int_to_peano(N, Kif).
kiertekel(Kif, Kif, true).

% int_to_peano(N, P): N termszetes szm Peano alakja P.
int_to_peano(0, 0).
int_to_peano(N, s(P)) :-
        N > 0, N1 is N-1,
        int_to_peano(N1, P).

% Peano szmok kirsnak formzsa
user:portray(Peano) :-
        peano_to_int(Peano, 0, N), write(N).
% felfggesztett clok kiratsnak formzsa
user:portray(user:Rel) :-
        Rel =.. [Op,A,B,C],
        (   Op = (+) ; Op = (-) ; Op = (*) ),
        Fun =.. [Op,A,B],
        print({Fun=C}).

% A Peano Peano-szm rtke N-N0.
peano_to_int(Peano, N0, N) :-
        nonvar(Peano),
        (   Peano == 0 -> N = N0
        ;   Peano = s(P), 
            N1 is N0+1,
            peano_to_int(P, N1, N)
        ).

user:goal_expansion({Korlat}, _, mini_nat:Cel) :-
	korlat_cel(Korlat, Cel).

% osszetett_cel(C1, C2, C12): A C12 cl a C1 s C2 clok konjunkcija.
osszetett_cel(true, Cel0, Cel) :- !,
	Cel = Cel0.
osszetett_cel(Cel0, true, Cel) :- !,
	Cel = Cel0.
osszetett_cel(Cel1, Cel2, (Cel1,Cel2)).
